@@ -31,22 +31,18 @@ from .batcher import batch_read_requests, batch_write_requests
 from .dist_store import get_or_create_store, LinearBarrier
 
 from .flatten import flatten, inflate
-from .io_preparer import (
-    _identity_tensor_prepare_func,
-    ObjectBufferConsumer,
-    prepare_read,
-    prepare_write,
-)
+from .io_preparer import prepare_read, prepare_write
 from .io_types import ReadIO, ReadReq, StoragePlugin, WriteIO, WriteReq
-from .knobs import get_is_batching_enabled
+from .knobs import is_batching_disabled
+
 from .manifest import (
     Entry,
-    get_available_entries,
-    is_replicated,
+    is_container_entry,
     Manifest,
     PrimitiveEntry,
     SnapshotMetadata,
 )
+from .manifest_ops import get_manifest_for_rank, handle_sharded_tensor_elasticity
 from .partitioner import consolidate_replicated_entries, partition_write_reqs
 from .pg_wrapper import PGWrapper
 from .rng_state import RNGState
@@ -69,102 +65,32 @@ T = TypeVar("T")
 
 class Snapshot:
     """
-    Snapshot represents the persisted program state at one point in time.
+    Create a reference to an existing snapshot.
 
-    Basic usage:
-    ::
+    Args:
+        path (str): The path to the snapshot. This should be the same as the
+            ``path`` argument used for :func:`Snapshot.take` when the snapshot
+            was taken.
 
-        # Define the program state
-        app_state = {"model": model, "optimizer": optimizer"}
+        pg (ProcessGroup, optional): The process group for the participants of
+            :meth:`Snapshot.restore`. If none, the default process group will be
+            used.
 
-        # At an appropriate time, persist the program state as a snapshot
-        snapshot = Snapshot.take(path=path, app_state=app_state)
-
-        # On resuming, restore the program state from a snapshot
-        snapshot.restore(app_state)
-
-    Overview:
-
-        At high level, torchsnapshot saves each value in state dicts as a
-        file/object in the corresponding storage system. It also saves a manifest
-        describing the persisted values and the structure of the original state
-        dict.
-
-        Comparing with :py:func:`torch.save` and :py:func:`torch.load`, torchsnapshot:
-
-        - Enables efficient random access of persisted model weights.
-
-        - Accelerates persistence by parallelizing writes.
-
-            - For replicated values, persistence is parallelized across ranks.
-
-        - Enables flexible yet robust elasticity (changing world size on
-          restore).
-
-
-    Elasticity:
-
-        Elasticity is implemented via correctly making persisted values
-        available to a newly joined rank, and having it correctly restores the
-        corresponding runtime objects from the persisted values.
-
-        For the purpose of elasticity, all persisted values fall into one of
-        the categories in [per-rank, replicated, sharded].
-
-        per-rank:
-
-            By default, all non-sharded values are treated as per-rank.
-
-            On save, the value is only saved by the owning rank.
-
-            On load, the value is only made available to the same rank.
-
-        replicated:
-
-            A user can suggest any non-sharded value as replicated via glob
-            patterns.
-
-            On save, the value is only saved once (can be by any rank).
-
-            On load, the value is made available to all ranks, including newly
-            joined ranks.
-
-        sharded:
-
-            Specific types are always treated as sharded (e.g. ShardedTensor).
-
-            On save, all shard-owning ranks save their shards.
-
-            On load, all shards are made available to all ranks, including
-            newly joined rank. All ranks can read from all shards for
-            restoring the runtime object from persisted values.
-            (ShardedTensor resharding is powered by torch.dist.checkpoint).
-
-        If all values within a snapshot are either replicated or sharded, the
-        snapshot is automatically reshard-able.
-
-        If a snapshot contains per-rank values, it cannot be resharded unless
-        the per-rank values are explicitly coerced to replicated on load.
+        storage_options (Dict[str, Any], optional): Additional keyword options
+            for the storage plugin to use. See each storage plugin's documentation
+            for customizations.
     """
 
     def __init__(
         self,
         path: str,
         pg: Optional[dist.ProcessGroup] = None,
+        storage_options: Optional[Dict[str, Any]] = None,
     ) -> None:
-        """
-        Initializes the reference to an existing snapshot.
-
-        Args:
-            path: The location of the snapshot.
-            pg: The process group for the processes restoring from the snapshot.
-                When unspecified:
-                    - If distributed is initialized, the global process group will be used.
-                    - If distributed is not initialized, single process is assumed.
-        """
         self.path: str = path
         self.pg: Optional[dist.ProcessGroup] = pg
         self._metadata: Optional[SnapshotMetadata] = None
+        self._storage_options = storage_options
 
     @classmethod
     def take(
@@ -173,23 +99,49 @@ class Snapshot:
         app_state: AppState,
         pg: Optional[dist.ProcessGroup] = None,
         replicated: Optional[List[str]] = None,
+        storage_options: Optional[Dict[str, Any]] = None,
         _custom_tensor_prepare_func: Optional[
             Callable[[str, torch.Tensor, bool], torch.Tensor]
         ] = None,
     ) -> "Snapshot":
         """
-        Take a snapshot from the program state.
+        Takes a snapshot of the application state.
 
         Args:
-            app_state: The program state to take the snapshot from.
-            path: The location to save the snapshot.
-            pg: The process group for the processes taking the snapshot.
-            When unspecified:
-                    - If distributed is initialized, the global process group will be used.
-                    - If distributed is not initialized, single process is assumed.
-            replicated: A list of glob patterns for hinting the matching paths
-                as replicated. Note that patterns not specified by all ranks
-                are ignored.
+            app_state (Dict[str, Stateful]): The application state to persist.
+                It takes the form of a dictionary, with the keys being
+                user-defined strings and the values being stateful objects.
+                Stateful objects are objects that exposes ``.state_dict()`` and
+                ``.load_state_dict()`` methods. Common PyTorch objects such as
+                :class:`torch.nn.Module`, :class:`torch.optim.Optimizer`, and
+                LR schedulers all qualify as stateful objects.
+
+            path (str): The location to save the snapshot. ``path`` can have a
+                URI prefix (e.g. ``s3://``) that specifies a storage backend.
+                If no URI prefix is supplied, ``path`` is assumed to be a file
+                system location. For distributed snapshot, if ``path`` is
+                inconsistent across participating ranks, the value specified by
+                rank 0 will be used. For multi-host snapshot, ``path`` needs to
+                be a location accessible by all hosts.
+
+                .. note:: ``path`` must **not** point to an existing snapshot.
+
+            pg (ProcessGroup, optional): The process group for the participants
+                of :meth:`Snapshot.take`. If none, the default process group will
+                be used.
+
+            replicated (List[str], optional): Glob patterns for marking
+                checkpoint content as replicated. Matching objects will be deduped
+                and load-balanced across ranks.
+
+                .. note:: The replication property is automatically inferred
+                    for ``DistributedDataParallel``. Only specify this argument
+                    if your model has fully replicated states but does not use
+                    ``DistributedDataParallel``.
+
+            storage_options (Dict[str, Any], optional): Additional keyword
+                options for the storage plugin to use. See each storage plugin's
+                documentation for customizations.
 
         Returns:
             The newly taken snapshot.
@@ -207,7 +159,7 @@ class Snapshot:
             replicated=replicated or [],
         )
         storage = url_to_storage_plugin_in_event_loop(
-            url_path=path, event_loop=event_loop
+            url_path=path, event_loop=event_loop, storage_options=storage_options
         )
         pending_io_work, metadata = cls._take_impl(
             path=path,
@@ -232,7 +184,7 @@ class Snapshot:
 
         storage.sync_close(event_loop=event_loop)
         event_loop.close()
-        snapshot = cls(path=path, pg=pg)
+        snapshot = cls(path=path, pg=pg, storage_options=storage_options)
         snapshot._metadata = metadata
         return snapshot
 
@@ -243,34 +195,29 @@ class Snapshot:
         app_state: AppState,
         pg: Optional[dist.ProcessGroup] = None,
         replicated: Optional[List[str]] = None,
+        storage_options: Optional[Dict[str, Any]] = None,
         _custom_tensor_prepare_func: Optional[
             Callable[[str, torch.Tensor, bool], torch.Tensor]
         ] = None,
     ) -> "PendingSnapshot":
         """
-        Asynchronously take a snapshot from the program state.
+        Asynchronously takes a snapshot from the application state.
 
-        This method creates a consistent snapshot of the app state (i.e.
-        changes to the app state after this method returns have no effect on
-        the snapshot). The asynchronicity is a result of performing storage I/O
-        in the background.
+        This function is identical to :func:`Snapshot.take`, except that it
+        returns early and performs as much I/O operations in the background as
+        possible, allowing training to resume early.
 
         Args:
-            app_state: The program state to take the snapshot from.
-            path: The location to save the snapshot.
-            pg: The process group for the processes taking the snapshot.
-            When unspecified:
-                    - If distributed is initialized, the global process group will be used.
-                    - If distributed is not initialized, single process is assumed.
-            replicated: A list of glob patterns for hinting the matching paths
-                as replicated. Note that patterns not specified by all ranks
-                are ignored.
+            app_state (Dict[str, Stateful]): Same as the ``app_state`` argument of :func:`Snapshot.take`.
+            path (str): Same as the ``path`` argument of :func:`Snapshot.take`.
+            pg (ProcessGroup, optional): Same as the ``pg`` argument of :func:`Snapshot.take`.
+            replicated (List[str], optional): Same as the ``replicated`` argument of :func:`Snapshot.take`.
+            storage_options (Dict[str, Any], optional): Same as the ``storage_options`` argument of :func:`Snapshot.take`.
 
         Returns:
-            A handle with which the newly taken snapshot can be obtained via
-            `.wait()`. Note that waiting on the handle is optional. The
-            snapshot will be committed regardless of whether `.wait()` is
-            invoked.
+            A handle to the pending snapshot. The handle has exposes a
+            ``.done()`` method for querying the progress and a ``.wait()``
+            method for waiting for the snapshot's completion.
         """
         torch._C._log_api_usage_once("torchsnapshot.Snapshot.async_take")
         cls._validate_app_state(app_state)
@@ -284,7 +231,7 @@ class Snapshot:
             replicated=replicated or [],
         )
         storage = url_to_storage_plugin_in_event_loop(
-            url_path=path, event_loop=event_loop
+            url_path=path, event_loop=event_loop, storage_options=storage_options
         )
 
         pending_io_work, metadata = cls._take_impl(
@@ -305,6 +252,7 @@ class Snapshot:
             metadata=metadata,
             storage=storage,
             event_loop=event_loop,
+            storage_options=storage_options,
         )
 
     @classmethod
@@ -325,9 +273,6 @@ class Snapshot:
         rng_state_item = cls._pop_rng_state(app_state=app_state)
         rng_state_dict = None
 
-        if not _custom_tensor_prepare_func:
-            _custom_tensor_prepare_func = _identity_tensor_prepare_func
-
         manifest: Manifest = {}
         flattened: Dict[str, Any] = {}
 
@@ -375,6 +320,7 @@ class Snapshot:
 
         object_entries: Dict[str, Entry] = {}
         logical_path_to_write_reqs: Dict[str, List[WriteReq]] = {}
+        primitive_entries: Dict[str, PrimitiveEntry] = {}
 
         for logical_path, obj in flattened.items():
             entry, wrs = prepare_write(
@@ -385,10 +331,17 @@ class Snapshot:
                 is_async_snapshot=is_async_snapshot,
                 _tensor_prepare_func=functools.partial(
                     _custom_tensor_prepare_func, logical_path
-                ),
+                )
+                if _custom_tensor_prepare_func is not None
+                else None,
             )
-            object_entries[logical_path] = entry
-            logical_path_to_write_reqs[logical_path] = wrs
+            # Primitive entries don't have write requests
+            # and don't need to be partitioned
+            if isinstance(entry, PrimitiveEntry):
+                primitive_entries[logical_path] = entry
+            else:
+                object_entries[logical_path] = entry
+                logical_path_to_write_reqs[logical_path] = wrs
 
         object_entries, logical_path_to_write_reqs = partition_write_reqs(
             entries=object_entries, write_reqs=logical_path_to_write_reqs, pg=pg_wrapper
@@ -397,15 +350,14 @@ class Snapshot:
             wr for wrs in logical_path_to_write_reqs.values() for wr in wrs
         ]
 
-        if get_is_batching_enabled():
-            entry_keys = list(object_entries.keys())
-            entries = list(object_entries.values())
-            entries, write_reqs = batch_write_requests(
-                entries=entries, write_reqs=write_reqs
+        if not is_batching_disabled():
+            _, write_reqs = batch_write_requests(
+                entries=list(object_entries.values()), write_reqs=write_reqs
             )
-            object_entries = dict(zip(entry_keys, entries))
 
-        manifest.update(object_entries)
+        all_entries = dict(**primitive_entries, **object_entries)
+
+        manifest.update(all_entries)
         manifest = cls._gather_manifest(manifest=manifest, pg=pg_wrapper)
 
         memory_budget_bytes = get_process_memory_budget_bytes(pg=pg_wrapper)
@@ -425,19 +377,23 @@ class Snapshot:
 
     def restore(self, app_state: AppState) -> None:
         """
-        Restores the program state from the snapshot.
+        Restores the application state from the snapshot.
 
         Args:
-            app_state: The program state to restore from the snapshot.
+            app_state (Dict[str, Stateful]): The application state to restore.
+                ``app_state`` needs to be either identical to or a subset of the
+                ``app_state`` used for :func:`Snapshot.take` when the snapshot was
+                taken.
         """
         torch._C._log_api_usage_once("torchsnapshot.Snapshot.restore")
         self._validate_app_state(app_state)
 
         event_loop = asyncio.new_event_loop()
         pg_wrapper = PGWrapper(self.pg)
-        rank = pg_wrapper.get_rank()
         storage = url_to_storage_plugin_in_event_loop(
-            url_path=self.path, event_loop=event_loop
+            url_path=self.path,
+            event_loop=event_loop,
+            storage_options=self._storage_options,
         )
 
         app_state = app_state.copy()
@@ -446,15 +402,10 @@ class Snapshot:
         global_keys = self._gather_keys(
             keys=list(app_state.keys()), pg_wrapper=pg_wrapper
         )
-        available_entries = get_available_entries(
-            manifest=self.metadata.manifest, rank=rank
-        )
         for key in global_keys:
             self._load_stateful(
-                rank=rank,
                 stateful_key=key,
                 stateful=app_state.get(key),
-                available_entries=available_entries,
                 storage=storage,
                 pg=pg_wrapper,
                 event_loop=event_loop,
@@ -465,10 +416,8 @@ class Snapshot:
         if rng_state_item is not None:
             key, stateful = rng_state_item
             self._load_stateful(
-                rank=rank,
                 stateful_key=key,
                 stateful=stateful,
-                available_entries=available_entries,
                 storage=storage,
                 pg=pg_wrapper,
                 event_loop=event_loop,
@@ -481,7 +430,9 @@ class Snapshot:
         if self._metadata is None:
             event_loop = asyncio.new_event_loop()
             storage = url_to_storage_plugin_in_event_loop(
-                url_path=self.path, event_loop=event_loop
+                url_path=self.path,
+                event_loop=event_loop,
+                storage_options=self._storage_options,
             )
             self._metadata = self._read_snapshot_metadata(
                 storage=storage, event_loop=event_loop
@@ -497,31 +448,25 @@ class Snapshot:
         memory_budget_bytes: Optional[int] = None,
     ) -> T:
         """
-        Read a persisted object from the snapshot's content.
-
-        The persisted object to read is specified by its path in the snapshot
-        metadata. Available paths can be obtained via `snapshot.get_manifest()`.
-
-        A path in snapshot metadata follows the following format:
-
-            ``RANK/STATEFUL_NAME/STATE_DICT_KEY[/NESTED_CONTAINER_KEY...]``
-
-        The rank only matters when the persisted object is "per-rank".
-        Arbitrary rank can be used when the persisted object is "replicated" or
-        "sharded".
-
-        If the persisted object is a sharded tensor, `obj_out` must be
-        supplied. The supplied tensor can be either a tensor or sharded tensor.
-        `read_object` will correctly populate `obj_out`'s data according to
-        sharding spec.
+        Reads an object from the snapshot's content.
 
         Args:
-            path: The path to the persisted object.
-            obj_out: If specified and the object type supports in-place load,
-                `read_object` will directly read the persisted object into
-                `obj_out`'s buffer.
-            memory_budget_bytes: When specified, the read operation will keep
-                the temporary memory buffer size below this threshold.
+            path (str): The path to the target object within the snapshot.
+                ``path`` is equivalent to the target object's key in the
+                snapshot manifest and can be obtained via
+                :meth:`Snapshot.get_manifest`.
+
+            obj_out (Any, optional): When specified, load the object in-place
+                into ``obj_out`` if in-place load is supported for the object's
+                type. Otherwise, ``obj_out`` is ignored.
+
+                .. note::
+                    When the target object is a ``ShardedTensor``, ``obj_out``
+                    must be specified.
+
+            memory_budget_bytes (int, optional): When specified, the read
+                operation will keep the temporary memory buffer size below this
+                threshold.
 
         Returns:
             The object read from the snapshot's content.
@@ -534,9 +479,11 @@ class Snapshot:
         # available to the rank (2) sharded tensor shards saved by all ranks
         # are made available to the rank. The availability of the entries is
         # determined from the perspective of the rank specified in the path.
-        manifest = get_available_entries(manifest=self.metadata.manifest, rank=rank)
+        manifest, merged_sd_entries = get_manifest_for_rank(
+            metadata=self.metadata, rank=rank
+        )
 
-        if unranked_path not in manifest:
+        if unranked_path not in merged_sd_entries and unranked_path not in manifest:
             # TODO: show candidates based on edit distance
             raise RuntimeError(
                 f'The supplied path "{path}" does not exist in the snapshot\'s manifest. '
@@ -551,12 +498,14 @@ class Snapshot:
         event_loop = asyncio.new_event_loop()
         pg_wrapper = PGWrapper(self.pg)
         storage = url_to_storage_plugin_in_event_loop(
-            url_path=self.path, event_loop=event_loop
+            url_path=self.path,
+            event_loop=event_loop,
+            storage_options=self._storage_options,
         )
-        entry = manifest[unranked_path]
+        entry = merged_sd_entries.get(unranked_path) or manifest[unranked_path]
         if isinstance(entry, PrimitiveEntry):
             return cast(T, entry.get_value())
-        read_reqs = prepare_read(
+        read_reqs, fut = prepare_read(
             entry=entry,
             obj_out=obj_out,
             # TODO: find a suitable buffer_size_limit_bytes to enable chunked
@@ -565,17 +514,8 @@ class Snapshot:
             # reading a single tensor.
             buffer_size_limit_bytes=memory_budget_bytes,
         )
-        box = []
-        for read_req in read_reqs:
-            buffer_consumer = read_req.buffer_consumer
-            if isinstance(buffer_consumer, ObjectBufferConsumer):
-                # ObjectBufferConsumer deals with objects that can not be
-                # in-place restored. We need to replace the original object
-                # in the flattened dictionary with the object materialized
-                # by the buffer consumer.
-                buffer_consumer.set_consume_callback(functools.partial(box.append))
 
-        if get_is_batching_enabled():
+        if not is_batching_disabled():
             read_reqs = batch_read_requests(read_reqs=read_reqs)
 
         sync_execute_read_reqs(
@@ -588,21 +528,19 @@ class Snapshot:
         )
         storage.sync_close(event_loop=event_loop)
         event_loop.close()
-        if len(box) != 0:
-            if len(box) != 1:
-                raise AssertionError(
-                    f"Expect to load a single object from an entry (got {len(box)})."
-                )
-            return box[0]
-        else:
-            return cast(T, obj_out)
+        return fut.obj
 
     def get_manifest(self) -> Dict[str, Entry]:
         """
-        Returns the snapshot's manifest.
+        Returns the snapshot manifest.
+
+        Each entry in the dictionary corresponds to an object in the snapshot,
+        with the keys being the logical paths to the objects and the values
+        being the metadata describing the object. For distributed snapshots,
+        the manifest contain entries for objects saved by all ranks.
 
         Returns:
-            The snapshot's manifest.
+            The snapshot manifest.
         """
         return copy.deepcopy(self.metadata.manifest)
 
@@ -651,13 +589,10 @@ class Snapshot:
                     f"Expected Stateful in app_state for key {key}, got {value_type}."
                 )
 
-    @classmethod
-    def _load_stateful(
-        cls,
-        rank: int,
+    def _load_stateful(  # noqa
+        self,
         stateful_key: str,
         stateful: Optional[Stateful],
-        available_entries: Manifest,
         storage: StoragePlugin,
         pg: PGWrapper,
         event_loop: asyncio.AbstractEventLoop,
@@ -665,62 +600,57 @@ class Snapshot:
         if stateful is None:
             return
 
-        # There are two ways to restore a stateful:
-        # 1. Reconstruct the state dict from storage and use it to call .load_state_dict()
-        # 2. Obtain the state dict via .state_dict(), restore its values from storage,
-        # then use it to call .load_state_dict()
-        #
-        # When .state_dict() returns references to the original tensors, #2 is
-        # more memory-efficient, because a tensor loaded from storage can be
-        # freed as soon as its value is copied to the original tensor.
-        state_dict = stateful.state_dict()
-        mnfst, flattened = flatten(state_dict, prefix=stateful_key)
-        del state_dict
+        manifest, merged_sd_entries = get_manifest_for_rank(
+            metadata=self.metadata, rank=pg.get_rank()
+        )
 
+        # In most cases (e.g. when the stateful is an nn.Module), the stateful
+        # has already allocated memory for its tensors. Materializing the
+        # persisted state dict and invoking .load_state_dict() would result in
+        # a memory footprint that is 2x the size of the stateful. We can reduce
+        # the memory footprint by exploiting the fact that most .state_dict()
+        # implementations return references to the internal tensors. By loading
+        # directly into the already allocated tensors and use them to construct
+        # a state dict for .load_state_dict(), we can eliminate an extra
+        # intermediate copy of the state. Even if the tensors in the state dict
+        # are copies of the internal tensors, this approach would not use more
+        # memory compared to the baseline.
+        _, flattened = flatten(stateful.state_dict(), prefix=stateful_key)
+        flattened = {
+            k: v
+            for k, v in flattened.items()
+            # ShardedTensor became a subclass of torch.Tensor since PyTorch
+            # 1.13. We can drop the check for ShardedTensor once PyTorch 1.12.1
+            # is no longer supported.
+            if isinstance(v, (torch.Tensor, ShardedTensor))
+        }
+
+        handle_sharded_tensor_elasticity(
+            manifest=manifest,
+            merged_sd_entries=merged_sd_entries,
+            tensor_requests=list(flattened.keys()),
+        )
+
+        container_entries = {}
         read_reqs: List[ReadReq] = []
-        for logical_path, obj in flattened.items():
-            if logical_path not in available_entries:
-                raise RuntimeError(
-                    f"""
-When restoring from the snapshot, stateful object "{stateful_key}" requested
-path "{logical_path}" which was not available to rank {rank}.
-
-- If the entry does not exist in the snapshot, it means that the state dict
-  entry was introduced after the snapshot was taken. To partially restore from
-  the snapshot, please explicitly ignore the state dict entries missing from
-  the snapshot.
-
-- If the entry exists in the snapshot, it could mean that the world size has
-  changed and the entry was not marked as replicated when the snapshot was
-  taken. To resolve the issue, try any of:
-    - Re-taking the snapshot with the new world size
-    - Re-taking the snapshot with the original world size, ensuring all
-          non-sharded values are marked as replicated
-    - Coerce the missing entry into replicated on restore"""
-                )
-
-            entry = available_entries[logical_path]
-            if isinstance(entry, PrimitiveEntry):
-                # for primitive types, directly materialize from PrimitiveEntry
-                flattened[logical_path] = entry.get_value()
+        futs = {}
+        for logical_path, entry in manifest.items():
+            if is_container_entry(entry):
+                container_entries[logical_path] = entry
                 continue
-            rrs = prepare_read(
-                entry=entry,
-                obj_out=obj,
-            )
-            for rr in rrs:
-                buffer_consumer = rr.buffer_consumer
-                if isinstance(buffer_consumer, ObjectBufferConsumer):
-                    # ObjectBufferConsumer deals with objects that can not be
-                    # in-place restored. We need to replace the original object
-                    # in the flattened dictionary with the object materialized
-                    # by the buffer consumer.
-                    buffer_consumer.set_consume_callback(
-                        functools.partial(dict.__setitem__, flattened, logical_path)
-                    )
-            read_reqs += rrs
 
-        if get_is_batching_enabled():
+            rrs, fut = prepare_read(
+                entry=entry,
+                obj_out=flattened.get(logical_path),
+            )
+            read_reqs += rrs
+            futs[logical_path] = fut
+
+            # Free memory in case the items is a copy
+            if logical_path in flattened:
+                del flattened[logical_path]
+
+        if not is_batching_disabled():
             read_reqs = batch_read_requests(read_reqs=read_reqs)
 
         memory_budget_bytes = get_process_memory_budget_bytes(pg=pg)
@@ -732,7 +662,12 @@ path "{logical_path}" which was not available to rank {rank}.
             event_loop=event_loop,
         )
 
-        state_dict = inflate(mnfst, flattened, prefix=stateful_key)
+        # Build the originally saved state dict and use it to restore the stateful
+        state_dict = inflate(
+            manifest=container_entries,
+            flattened={k: fut.obj for k, fut in futs.items()},
+            prefix=stateful_key,
+        )
         stateful.load_state_dict(state_dict)
 
     @staticmethod
@@ -853,17 +788,6 @@ path "{logical_path}" which was not available to rank {rank}.
         pg.all_gather_object(manifests, manifest)
         manifests = consolidate_replicated_entries(rank_to_entries=manifests)
 
-        # Remove replicated entries from non-zero ranks to reduce the size and
-        # serialization time of the yaml. In restore()/read_object(),
-        # replicated entries will be made available to all ranks via
-        # get_available_entries().
-        for rank, manifest in enumerate(manifests):
-            if rank == 0:
-                continue
-            manifests[rank] = {
-                k: v for k, v in manifest.items() if not is_replicated(v)
-            }
-
         global_manifest = {}
         for rank, manifest in enumerate(manifests):
             for logical_path, entry in manifest.items():
@@ -882,12 +806,14 @@ class PendingSnapshot:
         metadata: SnapshotMetadata,
         storage: StoragePlugin,
         event_loop: asyncio.AbstractEventLoop,
+        storage_options: Optional[Dict[str, Any]] = None,
     ) -> None:
         self.path = path
         self.pg: Optional[dist.ProcessGroup] = pg_wrapper.pg
         # pyre-ignore
         self.exc_info: Optional[Any] = None
         self._done = False
+        self._storage_options = storage_options
 
         self.thread = Thread(
             target=self._complete_snapshot,
@@ -955,7 +881,9 @@ class PendingSnapshot:
             raise RuntimeError(
                 f"Encountered exception while taking snapshot asynchronously:\n{formatted}"
             )
-        return Snapshot(path=self.path, pg=self.pg)
+        return Snapshot(
+            path=self.path, pg=self.pg, storage_options=self._storage_options
+        )
 
     def done(self) -> bool:
         return self._done
